/*
 *                      Yeppp! library build framework
 *
 * This file is part of Yeppp! library infrastructure and licensed under 2-clause BSD license.
 * See library/LICENSE.txt for details.
 *
 */

package info.yeppp.Build;

import info.yeppp.Build.internals.*;
import info.yeppp.Build.internals.Process;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.SystemUtils;

import java.io.File;
import java.util.HashMap;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class GnuAssembler {
	public GnuAssembler() {
		assemblerPath = "as";
	}

	public BuildMessages assemble(final String sourcePath) {
		if (sourceDirectory == null) {
			throw new IllegalStateException("Source directory was not assigned any value");
		}
		if (objectDirectory == null) {
			throw new IllegalStateException("Object directory was not assigned any value");
		}
		final BuildMessages buildMessages = new BuildMessages();
		final String relativeSourcePath = FileSystem.getRelativePath(sourcePath, sourceDirectory);
		if (relativeSourcePath == null) {
			throw new IllegalArgumentException("Source path is not within source directory");
		}
		final String objectPath = FilenameUtils.concat(objectDirectory,
				FilenameUtils.removeExtension(relativeSourcePath) + (SystemUtils.IS_OS_WINDOWS ? ".obj" : ".o"));

		prepareDirectories(objectPath);
		final info.yeppp.Build.internals.Process.Parameters parameters = prepareParameters(sourcePath, objectPath);
		final Process.Environment environment = new Process.Environment();
		final Process.EventHandler eventHandler = new Process.EventHandler() {
			@Override
			public void onStdOutLineWrite(String line) {
				System.out.println(line);
			}

			@Override
			public void onStdErrLineWrite(String line) {
				System.err.println(line);
			}

			@Override
			public void onError(String description) {
				buildMessages.add(new BuildMessage(BuildMessage.Type.Exception, sourcePath, null, null, description));
			}

			@Override
			public void onCompletion(Integer status) {
			}
		};
		final Process assembler = new Process(assemblerPath, parameters, environment, eventHandler);
		assembler.join();
		return buildMessages;
	}

	protected Process.Parameters prepareParameters(String sourcePath, String objectPath) {
		final Process.Parameters parameters = new Process.Parameters();
		parameters.add("-o" + objectPath, sourcePath);
		return parameters;
	}

	public String getObjectPath(String sourcePath) {
		if (sourceDirectory == null) {
			throw new IllegalStateException("Source directory was not assigned any value");
		}
		if (objectDirectory == null) {
			throw new IllegalStateException("Object directory was not assigned any value");
		}
		final String relativeSourcePath = FileSystem.getRelativePath(sourcePath, sourceDirectory);
		if (relativeSourcePath == null) {
			throw new IllegalArgumentException("Source path is not within source directory");
		}
		return FilenameUtils.concat(objectDirectory,
				FilenameUtils.removeExtension(relativeSourcePath) + (SystemUtils.IS_OS_WINDOWS ? ".obj" : ".o"));
	}

	public void setSourceDirectory(String sourceDirectory) {
		this.sourceDirectory = sourceDirectory;
	}

	public String getSourceDirectory() {
		return this.sourceDirectory;
	}

	public void setObjectDirectory(String objectDirectory) {
		this.objectDirectory = objectDirectory;
	}

	public String getObjectDirectory() {
		return this.objectDirectory;
	}

	public boolean addMacro(String name, String value) {
		if (macros.containsKey(name)) {
			return false;
		} else {
			macros.put(name, value);
			return true;
		}
	}

	public boolean removeMacro(String name) {
		return (macros.remove(name) != null);
	}

	public void clearMacros() {
		macros.clear();
	}

	private void prepareDirectories(String path) {
		File file = new File(path);
		if (!file.isDirectory()) {
			file = file.getParentFile();
		}
		file.mkdirs();
	}

	private String installPath = null;
	private String assemblerPath = null;
	private String sourceDirectory = null;
	private String objectDirectory = null;
	private Map<String, String> macros = new HashMap<String, String>();
}
