/*
 *                      Yeppp! library build framework
 *
 * This file is part of Yeppp! library infrastructure and licensed under 2-clause BSD license.
 * See library/LICENSE.txt for details.
 *
 */

package info.yeppp.Build;

import info.yeppp.Build.internals.CommandLineTool;
import info.yeppp.Build.internals.FileSystem;
import info.yeppp.Build.sdk.Windows8Sdk;
import info.yeppp.Build.sdk.WindowsSdk;
import org.apache.commons.io.FileUtils;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.SystemUtils;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class MicrosoftCppCompiler extends AbstractCppCompiler {

	public MicrosoftCppCompiler(VisualStudio visualStudio, WindowsSdk windowsSdk, Architecture architecture) throws Exception {
		if (!visualStudio.isSupportedArchitecture(architecture)) {
			throw new UnsupportedOperationException(String.format("Architecture %s is not supported in Visual Studio 2010", architecture.toString()));
		}
		this.architecture = architecture;
		this.visualStudio = visualStudio;
		this.windowsSdk = windowsSdk;
		this.compilerPath = FilenameUtils.concat(visualStudio.getCppToolsPath(architecture), "cl.exe");
	}

	public BuildMessages compile(String sourcePath) {
		if (sourceDirectory == null) {
			throw new IllegalStateException("Source directory was not assigned any value");
		}
		if (objectDirectory == null) {
			throw new IllegalStateException("Object directory was not assigned any value");
		}
		final BuildMessages buildMessages = new BuildMessages();
		final String relativeSourcePath = FileSystem.getRelativePath(sourcePath, sourceDirectory);
		if (relativeSourcePath == null) {
			throw new IllegalArgumentException("Source path is not within source directory");
		}
		final String objectPath = FilenameUtils.concat(objectDirectory,
				FilenameUtils.removeExtension(relativeSourcePath) + (SystemUtils.IS_OS_WINDOWS ? ".obj" : ".o"));

		createDirectories(objectPath);
		FileUtils.deleteQuietly(new File(objectPath));


		try {
			final List<String> commandLine = composeCommandLine(sourcePath, objectPath);
			final Map<String, String> environment = composeEnvironment();
//				System.out.println(StringUtils.join(commandLine, " "));
			final CommandLineTool compiler = new CommandLineTool(commandLine, environment);
			for (String line : compiler.getError()) {
				System.err.println(line);
			}
			List<String> output = compiler.getOutput();
			File sourceFile = new File(sourcePath);
			if (output.get(0).equals(sourceFile.getName())) {
				output.remove(0);
			}
			buildMessages.add(parseCommandLineOutput(compiler.getOutput()));
		} catch (IOException exception) {
			buildMessages.add(new BuildMessage(BuildMessage.Type.Exception, relativeSourcePath, null, null, exception.getMessage()));
		} catch (Exception e) {
			e.printStackTrace();
		}
		return buildMessages;
	}

	protected List<String> composeCommandLine(String sourcePath, String objectPath) {
		final List<String> commandLine = new ArrayList<String>();
		commandLine.add(compilerPath);
		commandLine.add("/nologo");
		commandLine.add("/c");
		commandLine.add("/Fo" + objectPath);
		if (debugEnabled) {
			commandLine.add("/GS");
			commandLine.add("/Zi");
		} else {
			commandLine.add("/GS-");
		}
		for (String include : includes) {
			commandLine.add("/I" + include);
		}
		if (debugEnabled) {
			commandLine.add("/Od");
		} else {
			switch (optimization) {
				case NoOptimization:
					commandLine.add("/Od"); break;
				case MinSizeOptimization:
					commandLine.add("/O1"); break;
				case SpeedOptimization:
					commandLine.add("/O2"); break;
				case MaxSpeedOptimization:
					commandLine.add("/Ox"); break;
			}
		}
		for (Map.Entry<String, String> macro: macros.entrySet()) {
			commandLine.add("/D" + macro.getKey() + "=" + macro.getValue());
		}
		switch (runtimeLibrary) {
			case NoRuntimeLibrary:
				commandLine.add("/Zl"); break;
			case DynamicRuntimeLibrary:
				commandLine.add(debugEnabled ? "/MDd" : "/MD"); break;
			case StaticRuntimeLibrary:
				commandLine.add(debugEnabled ? "/MTd" : "/MT"); break;
		}
		commandLine.add(rttiEnabled ? "/GR" : "/GR-");
		switch (exceptions) {
			case NoExceptions:
				commandLine.add("/EHs-"); break;
			case DisabledExceptions:
				commandLine.add("/EHs-"); break;
			case SyncronousExceptions:
				commandLine.add("/EHsc"); break;
			case AsyncronousExceptions:
				commandLine.add("/EHa"); break;
		}
		commandLine.add(sourcePath);
		return commandLine;
	}

	protected BuildMessages parseCommandLineOutput(List<String> lines) {
		final BuildMessages buildMessages = new BuildMessages();
		final Pattern errorPattern = Pattern.compile("^(.+)\\((\\d+)\\): error C(\\d+): (.+)$");
		final Pattern fatalErrorPattern = Pattern.compile("(.+)\\((\\d+)\\): fatal error C(\\d+): (.+)$");
		final Pattern warningPattern = Pattern.compile("^(.+)\\((\\d+)\\): warning C(\\d+): (.+)$");
		Matcher matcher;
		for (String line : lines) {
			if ((matcher = errorPattern.matcher(line)).find()) {
				final String errorPath = FileSystem.getRelativePath(matcher.group(1), this.sourceDirectory);
				final int lineNumber = Integer.parseInt(matcher.group(2));
				final int errorNumber = Integer.parseInt(matcher.group(3));
				final String errorMessage = matcher.group(4);
				buildMessages.add(new BuildMessage(BuildMessage.Type.Error, errorPath, lineNumber, errorNumber, errorMessage));
			} else if ((matcher = fatalErrorPattern.matcher(line)).find()) {
				final String errorPath = FileSystem.getRelativePath(matcher.group(1), this.sourceDirectory);
				final int lineNumber = Integer.parseInt(matcher.group(2));
				final int errorNumber = Integer.parseInt(matcher.group(3));
				final String errorMessage = matcher.group(4);
				buildMessages.add(new BuildMessage(BuildMessage.Type.FatalError, errorPath, lineNumber, errorNumber, errorMessage));
			} else if ((matcher = warningPattern.matcher(line)).find()) {
				final String warningPath = FileSystem.getRelativePath(matcher.group(1), this.sourceDirectory);
				final int lineNumber = Integer.parseInt(matcher.group(2));
				final int warningNumber = Integer.parseInt(matcher.group(3));
				final String warningMessage = matcher.group(4);
				buildMessages.add(new BuildMessage(BuildMessage.Type.Warning, warningPath, lineNumber, warningNumber, warningMessage));
			} else {
				final BuildMessage lastBuildMessage = buildMessages.getLastBuildMessage();
				if (lastBuildMessage == null) {
					buildMessages.add(new BuildMessage(BuildMessage.Type.Error, null, null, null, line));
				} else {
					lastBuildMessage.appendMessage(SystemUtils.LINE_SEPARATOR + line);
				}
			}
		}
		return buildMessages;
	}

	private Map<String, String> composeEnvironment() throws IOException {
		final Map<String, String> environment = new HashMap<String, String>(System.getenv());
		boolean processedIncludeVariable = false;
		for (Map.Entry<String, String> variable : environment.entrySet()) {
			if (variable.getKey().equalsIgnoreCase("PATH")) {
				String path = variable.getValue();
				path = visualStudio.getIdePath() +
						SystemUtils.PATH_SEPARATOR + visualStudio.getCommonToolsPath() +
						SystemUtils.PATH_SEPARATOR + visualStudio.getCppToolsPath(architecture) +
						SystemUtils.PATH_SEPARATOR + path;
				variable.setValue(path);
			} else if (variable.getKey().equalsIgnoreCase("INCLUDE")) {
				String include = variable.getValue();
				include = StringUtils.join(globalIncludes, SystemUtils.PATH_SEPARATOR) +
						SystemUtils.PATH_SEPARATOR + include;
				variable.setValue(include);
				processedIncludeVariable = true;
			}
		}
		if (processedIncludeVariable == false) {
			environment.put("INCLUDE", StringUtils.join(globalIncludes, SystemUtils.PATH_SEPARATOR));
		}
		return environment;
	}

	private void createDirectories(String path) {
		File file = new File(path);
		if (!file.isDirectory()) {
			file = file.getParentFile();
		}
		file.mkdirs();
	}

	@Override
	public void addDefaultGlobalIncludeDirectories() {
		this.addGlobalIncludeDirectory(visualStudio.getCppIncludePath());
		this.addGlobalIncludeDirectory(windowsSdk.getIncludePath());
		if (windowsSdk instanceof Windows8Sdk) {
			this.addGlobalIncludeDirectory(((Windows8Sdk) windowsSdk).getSharedIncludePath());
		}
	}

	private final Architecture architecture;
	private final VisualStudio visualStudio;
	private final WindowsSdk windowsSdk;
	private final String compilerPath;

}
